;;; x-pgp-sig.el -- Yet another PGP authenticating utilities for Emacsen.

;;; Copyright (C) 1996, 1997 Katsumi Yamaoka
;;; Copyright (C) 1996, 1997 OKUNISHI Fujikazu
;;; Author:     Katsumi Yamaoka <yamaoka@ga.sony.co.jp>
;;;             OKUNISHI Fujikazu <fuji0924@mbox.kyoto-inet.or.jp>
;;; Maintainer: Katsumi Yamaoka <yamaoka@ga.sony.co.jp>
;;; Created:    97/05/29
;;; Version:    1.2.5
;;; References: ftp://ftp.uu.net/networking/news/misc/pgpcontrol/README.html
;;;             ftp://ftp.uu.net/networking/news/misc/pgpcontrol/README
;;; Keywords:   gnus, mh-e, mew, pgp, pgpverify, signcontrol, x-pgp-sig, OS/2

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with GNU Emacs; see the file COPYING.  If not, write to the
;;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;;; Boston, MA 02111-1307, USA.

;;; Commands:
;;;  x-pgp-sig-set-signer
;;;     "Set PGP signer and passphrase."
;;;  x-pgp-sig-set-signer-maybe
;;;     "Set PGP signer and passphrase if necessary."
;;;  x-pgp-sig-switch-signing
;;;     "Judging whether to add X-PGP signature."
;;;  x-pgp-sig-verify
;;;     "Verify X-PGP signature in current message."

;;; Functions:
;;;  x-pgp-sig-sign
;;;     "Add X-PGP signature."
;;--------------------------------------------------------------------

;;; Code:

(or (boundp 'running-xemacs)
    (defconst running-xemacs (string-match "XEmacs\\|Lucid" emacs-version)))

(defconst x-pgp-sig-version-number "1.2.5"
  "Version number for this version of x-pgp-sig. [x-pgp-sig.el]")
(defconst x-pgp-sig-version
  (format "Yet another PGP authenticating utilities v%s"
	  x-pgp-sig-version-number)
  "Version string for this version of x-pgp-sig. [x-pgp-sig.el]")

(require 'cl)

(defvar x-pgp-sig-from
  (or
   (and (boundp 'user-mail-address)
	(stringp user-mail-address)
	user-mail-address)
   (and (boundp 'mail-host-address)
	(stringp mail-host-address)
	(concat (user-login-name) "@" mail-host-address))
   (and (boundp 'gnus-local-domain)
	(stringp gnus-local-domain)
	(concat (user-login-name) "@" gnus-local-domain))
   "nobody@nowhere")
  "*Signer's mail address. [x-pgp-sig.el]")

(defvar x-pgp-sig-default-signer
  (format "%s <%s>" (user-full-name) x-pgp-sig-from)
  "*PGP user ID. [x-pgp-sig.el]")

(defvar x-pgp-sig-pgp-program "pgp" "*PGP executable. [x-pgp-sig.el]")

(defvar x-pgp-sig-toMime-program "toMime"
  "*toMime executable. [x-pgp-sig.el]")
(defvar x-pgp-sig-toMime-options nil
  "*List of toMime options. [x-pgp-sig.el]")

(defvar x-pgp-sig-mew-encode-message-header-function
  'x-pgp-sig-mew-encode-message-header-with-toMime
  "*Function to call to encode the message of Mew. Legal values include
`x-pgp-sig-mew-encode-message-header-with-toMime' (the default),
`x-pgp-sig-mew-encode-message-header-with-semi' and
`x-pgp-sig-mew-encode-message-header-with-tm'. [x-pgp-sig.el]")

(defvar x-pgp-sig-process-coding-system-for-output nil
  "*Coding system for PGP process (output from Mule/XEmacs). [x-pgp-sig.el]")
(defvar x-pgp-sig-process-kanji-code 2
  "*Kanji code for PGP process (NEmacs).
0:No-conversion will not work well. [x-pgp-sig.el]")

(defvar x-pgp-sig-pgppath
  (or (getenv "PGPPATH") (expand-file-name "~/.pgp"))
  "*Environment variable \"PGPPATH\". [x-pgp-sig.el]")

(defvar x-pgp-sig-pubring
  (or (getenv "PUBRING") (concat x-pgp-sig-pgppath "/pubring.pgp"))
  "*Environment variable \"PUBRING\". [x-pgp-sig.el]")

(defvar x-pgp-sig-secring
  (or (getenv "SECRING") (concat x-pgp-sig-pgppath "/secring.pgp"))
  "*Environment variable \"SECRING\". [x-pgp-sig.el]")

(defvar x-pgp-sig-delete-last-empty-line-while-verifying t
  "*If non-nil, delete last empty line of message while verifying.
\[x-pgp-sig.el]")

(defvar x-pgp-sig-add-content-length-field t
  "*If non-nil, add Content-Length field when signing. [x-pgp-sig.el]")

(defvar x-pgp-sig-approved-field-body x-pgp-sig-from
  "*Address for Approved field. [x-pgp-sig.el]")

(defvar x-pgp-sig-x-info-field-body nil
  "*X-Info field body. [x-pgp-sig.el]")

(defvar x-pgp-sig-sign-fields
  '("Subject" "Control" "Message-ID" "Date" "From" "Sender"
    "X-Mailer" "X-Newsreader" "X-Face")
  "*Fields to sign. [x-pgp-sig.el]")

(defvar x-pgp-sig-adjust-sign-fields-alist nil
  "*Adjust `x-pgp-sig-sign-fields' if matched text exists in narrowed header.

\((REGEXP1 (sub \"FIELD1\" \"FIELD2\" ...)
\          (add \"FIELD3\" \"FIELD4\" ...))
\ (REGEXP2 (sub \"FIELD5\" \"FIELD6\" ...)
\          (add \"FIELD7\" \"FIELD8\" ...)))

REGEXP can be a function as well. [x-pgp-sig.el]")

(defvar x-pgp-sig-last-fields '("X-Face" "X-Face-Version")
  "*PGP signed field will be added above these fields. [x-pgp-sig.el]")

(defvar x-pgp-sig-always-sign nil
  "*If non-nil, always add X-PGP signature. [x-pgp-sig.el]")

(defvar x-pgp-sig-add-signature-judging-function
  'x-pgp-sig-add-signature-y-or-n-p
  "*Function whether to add X-PGP signature. [x-pgp-sig.el]")

(defvar x-pgp-sig-load-hook nil
  "*Hook to be run after the x-pgp-sig package has been loaded.
\[x-pgp-sig.el]")

(defvar x-pgp-sig-sign-hook nil
  "*Hook run after adding X-PGP-Sig field. [x-pgp-sig.el]")

(defvar x-pgp-sig-prepare-sign-hook nil
  "*Hook run before signing pgp process. [x-pgp-sig.el]")

(defvar x-pgp-sig-post-sign-hook nil
  "*Hook run after signing pgp process. [x-pgp-sig.el]")

(defvar x-pgp-sig-good-signature-regexp "^Good signature from user.*$"
  "Regular expression matching a PGP signature validation message.
\[x-pgp-sig.el]")

(defvar x-pgp-sig-bad-signature-regexp "^Bad signature from user.*$"
  "Regular expression not matching a PGP signature validation message.
\[x-pgp-sig.el]")

(defvar x-pgp-sig-expected-re "Key ID \\(\\S +\\) not found")

(defvar x-pgp-sig-verify-method-alist
  '(
    (gnus-summary-mode
     . (gnus-summary-show-article
	x-pgp-sig-find-gnus-article-buffer
	gnus-summary-show-article))
    (gnus-Subject-mode
     . (gnus-Subject-show-article
	x-pgp-sig-find-gnus-article-buffer
	gnus-Subject-show-article))
    (mh-folder-mode
     . (x-pgp-sig-mh-get-article
	x-pgp-sig-msg-to-verify-buffer))
    (mew-summary-mode
     . (x-pgp-sig-mew-get-article
	x-pgp-sig-msg-to-verify-buffer))
    (mew-virtual-mode
     . (x-pgp-sig-mew-get-article
	x-pgp-sig-msg-to-verify-buffer))
    )
  "*Alist for verifying signature. [x-pgp-sig.el]
Each cells are:
'(MAJOR-MODE . (PREPARE-FUNCTION BUFFERorFUNCTION POST-FUNCTION))")

(defvar x-pgp-sig-*noconv* (cond (running-xemacs 'binary)
				 ((boundp 'MULE) *noconv*)))

;;; Internal variables

(defconst x-pgp-sig-field-name "X-PGP-Sig"
  "Field name of PGP signature. [x-pgp-sig.el]")

(defconst x-pgp-sig-signed-headers-field-name "X-Signed-Headers"
  "Field name of PGP signed fields. [x-pgp-sig.el]")

(defconst x-pgp-sig-msg-begin-line "-----BEGIN PGP MESSAGE-----"
  "Text for start of PGP message delimiter. [x-pgp-sig.el]")

(defconst x-pgp-sig-msg-end-line "-----END PGP MESSAGE-----"
  "Text for end of PGP message delimiter. [x-pgp-sig.el]")

(defconst x-pgp-sig-signed-msg-begin-line "-----BEGIN PGP SIGNED MESSAGE-----"
  "Text for start of PGP signed messages. [x-pgp-sig.el]")

(defconst x-pgp-sig-signed-begin-line "-----BEGIN PGP SIGNATURE-----"
  "Text for start of PGP signature. [x-pgp-sig.el]")

(defconst x-pgp-sig-signed-end-line "-----END PGP SIGNATURE-----"
  "Text for end of PGP signature. [x-pgp-sig.el]")

(defvar x-pgp-sig-working-buffer " *x-pgp-sig-working*")

(defvar x-pgp-sig-msg-to-verify-buffer nil)

(defvar x-pgp-sig-enable-sign)
(setq x-pgp-sig-enable-sign x-pgp-sig-always-sign)

(or (stringp x-pgp-sig-default-signer)
    (setq x-pgp-sig-default-signer
	  (format "%s <%s>" (user-full-name) x-pgp-sig-from)))

(defvar x-pgp-sig-current-signer)
(setq x-pgp-sig-current-signer x-pgp-sig-default-signer)

(defvar x-pgp-sig-passphrase-alist)
(setq x-pgp-sig-passphrase-alist nil)

(defvar x-pgp-sig-public-key-alist)
(setq x-pgp-sig-public-key-alist nil)

(defvar x-pgp-sig-secret-key-alist)
(setq x-pgp-sig-secret-key-alist nil)

(defconst x-pgp-sig-idea-passphrase
  (let ((i 256)
	p r)
    (while (> i 0)
      (decf i)
      (setq r (abs (random))
	    p
	    (concat
	     p
	     (char-to-string
	      (+ (/ (* 95 (- r (* (/ r 100) 100))) 100) 32)))))
    p))


;;; Emulations for each Emacs variants.

(and (boundp 'to-kanji-display)
     (setq to-ascii-display ?B
	   to-ascii-fileio ?B
	   to-ascii-process ?B
	   to-kanji-display ?B
	   to-kanji-fileio ?B
	   to-kanji-process ?B))

(defun x-pgp-sig-read-noecho (prompt)
  "Read a single line of text from user without echoing, and return it.
\[x-pgp-sig.el]"
  (let ((ans "")
	(c 0)
	(echo-keystrokes 0)
	(cursor-in-echo-area t)
	(message-log-max nil)
	(done nil))
    (while (not done)
      (message "%s" prompt)
      (setq c (read-char))
      (cond ((= c ?\C-g)
	     (setq quit-flag t
		   done t))
	    ((or (= c ?\r) (= c ?\n) (= c ?\e))
	     (setq done t))
	    ((= c ?\C-u)
	     (setq ans ""))
	    ((and (/= c ?\b) (/= c ?\177))
	     (setq ans (concat ans (char-to-string c))))
	    ((> (length ans) 0)
	     (setq ans (substring ans 0 -1)))))
    (if quit-flag
	(prog1
	    (setq quit-flag nil)
	  (message "Quit")
	  (beep t))
      (message "")
      ans)))

(fset 'x-pgp-sig-buffer-live-p
      (or (and (fboundp 'gnus-buffer-live-p)
	       (symbol-function 'gnus-buffer-live-p))
	  (function
	   (lambda (object)
	     "Say whether BUFFER is alive or not. [x-pgp-sig.el]"
	     (and object
		  (get-buffer object)
		  (buffer-name (get-buffer object))
		  )))))

(or (fboundp 'buffer-substring-no-properties)
    (fset 'buffer-substring-no-properties
	  (symbol-function 'buffer-substring)))

(or (fboundp 'member)
    (defun member (elt list)
      "Return non-nil if ELT is an element of LIST. Comparison done with EQUAL.
The value is actually the tail of LIST whose car is ELT. [x-pgp-sig.el]"
      (while (and list (not (equal elt (car list))))
	(setq list (cdr list)))
      list))

(or (fboundp 'put-alist)
    ;;; Imported from tl-list.el.
    (defun put-alist (item value alist)
      "Modify ALIST to set VALUE to ITEM.
If there is a pair whose car is ITEM, replace its cdr by VALUE.
If there is not such pair, create new pair (ITEM . VALUE) and
return new alist whose car is the new pair and cdr is ALIST. [x-pgp-sig.el]"
      (let ((pair (assoc item alist)))
	(if pair
	    (progn
	      (setcdr pair value)
	      alist)
	  (cons (cons item value) alist)
	  ))))

(or (fboundp 'delete)
    ;;; Imported from emu-18.el
    (defun delete (elt list)
      "Delete by side effect any occurrences of ELT as a member of LIST.
The modified LIST is returned.  Comparison is done with `equal'.
If the first member of LIST is ELT, deleting it is not a side effect;
it is simply using a different list.
Therefore, write `(setq foo (delete element foo))'
to be sure of changing the value of `foo'. [x-pgp-sig.el]"
      (if (equal elt (car list))
	  (cdr list)
	(let ((rest list)
	      (rrest (cdr list))
	      )
	  (while (and rrest (not (equal elt (car rrest))))
	    (setq rest rrest
		  rrest (cdr rrest))
	    )
	  (rplacd rest (cdr rrest))
	  list))))

(or (fboundp 'match-string)
    (defun match-string (num &optional string)
      "Return string of text matched by last search.
NUM specifies which parenthesized expression in the last regexp.
Value is nil if NUMth pair didn't match, or there were less than NUM pairs.
Zero means the entire text matched by the whole regexp or whole string.
STRING should be given if the last search was by `string-match' on STRING.
\[x-pgp-sig.el]"
      (if (match-beginning num)
	  (if string
	      (substring string (match-beginning num) (match-end num))
	    (buffer-substring (match-beginning num) (match-end num))))))


;;; Functions.

(defun x-pgp-sig-narrow-to-header (&optional include-separator)
  (widen)
  (goto-char (point-min))
  (narrow-to-region
   (point)
   (or (and (re-search-forward
	     (format "^\\(%s\\)?\n"
		     (regexp-quote mail-header-separator))
	     nil t)
	    (if include-separator
		(match-end 0)
	      (match-beginning 0)))
       (point-max)))
  (goto-char (point-min)))

(defun x-pgp-sig-process-buffer (args &optional passphrase buffer)
  (let ((obuf (or (and (x-pgp-sig-buffer-live-p buffer) buffer)
		  (current-buffer)))
	(pgpbuf (get-buffer-create " *PGP*"))
	(process-environment process-environment)
	process-connection-type case-fold-search envs env str proc err)
    (mapcar
     (function (lambda (a) (or (member a args) (setq args (cons a args)))))
     '("+verbose=1" "+language=en" "+batchmode"))
    (save-excursion
      (set-buffer obuf)
      (setq str (buffer-string))
      (set-buffer pgpbuf)
      (setq buffer-read-only nil)
      (erase-buffer)
      (setq case-fold-search nil)
      (if (fboundp 'setenv)
	  (setenv "PGPPASSFD" (and passphrase "0"))
	(setq envs process-environment)
	(while (setq env (pop envs))
	  (and (string-match "^PGPPASSFD=" env)
	       (setq process-environment (delete env process-environment))))
	(and passphrase
	     (setq process-environment
		   (cons "PGPPASSFD=0" process-environment))))
      (let (; NEmacs
	    (default-kanji-process-code
	      (or x-pgp-sig-process-kanji-code 0))
	    program-kanji-code-alist
	    ; Mule-{1,2}.x
	    (default-process-coding-system
	      (cons x-pgp-sig-*noconv*
		    (or x-pgp-sig-process-coding-system-for-output
			x-pgp-sig-*noconv*)))
	    program-coding-system-alist
	    ; >= XEmacs-20.1b7
	    (buffer-file-coding-system
	     (or x-pgp-sig-process-coding-system-for-output
		 x-pgp-sig-*noconv*)))
	(and passphrase (insert passphrase "\n"))
	(insert str)
	(apply 'call-process-region (point-min) (point-max)
	       x-pgp-sig-pgp-program t pgpbuf nil args))
      ;;; for 'OS/2 convert CRLF to LF
      (goto-char (point-min))
      (while (search-forward "\r\n" nil t)
	(replace-match "\n"))
      ;;;
      (goto-char (point-min))
      (setq case-fold-search t)
      (and (or (re-search-forward "error:[\t ]*\\(.*\\)$" nil t)
	       (re-search-forward
		"^\\(process .+ exited abnormally with code .+\\)$" nil t))
	   (setq err (match-string 1)))
      (setq str (buffer-string))
      (set-buffer obuf)
      (setq buffer-read-only nil)
      (erase-buffer)
      (insert str))
    (kill-buffer pgpbuf)
    err))

(defun x-pgp-sig-idea-encrypt-string (str)
  "Encrypt string using IDEA method. [x-pgp-sig.el]"
  (and (stringp str)
       (let ((buffer (get-buffer-create x-pgp-sig-working-buffer))
	     beg result)
	 (unwind-protect
	     (save-excursion
	       (set-buffer buffer)
	       (setq buffer-read-only nil)
	       (erase-buffer)
	       (insert str)
	       (setq str nil)
	       (setq result (x-pgp-sig-process-buffer
			     '("-fact") x-pgp-sig-idea-passphrase))
	       (and result (error "!! %s" result))
	       (goto-char (point-min))
	       (or (re-search-forward
		    (format "%s$" (regexp-quote x-pgp-sig-msg-begin-line))
		    nil t)
		   (error "!! Encryption failed."))
	       (or (search-forward "\n\n" nil t)
		   (error "!! Encryption failed."))
	       (setq beg (point))
	       (or (re-search-forward
		    (format "%s$" (regexp-quote x-pgp-sig-msg-end-line))
		    nil t)
		   (error "!! Encryption failed."))
	       (setq result (buffer-substring beg (match-beginning 0))))
	   (kill-buffer buffer))
	 result)))

(defun x-pgp-sig-idea-decrypt-string (str)
  "Decrypt string using IDEA method. [x-pgp-sig.el]"
  (and (stringp str)
       (let ((buffer (get-buffer-create x-pgp-sig-working-buffer))
	     result)
	 (unwind-protect
	     (save-excursion
	       (set-buffer buffer)
	       (setq buffer-read-only nil)
	       (erase-buffer)
	       (insert x-pgp-sig-msg-begin-line "\n\n" str)
	       (setq str nil)
	       (or (eolp) (insert "\n"))
	       (insert x-pgp-sig-msg-end-line "\n")
	       (setq result
		     (x-pgp-sig-process-buffer
		      '("-f") x-pgp-sig-idea-passphrase))
	       (and result (error "!! %s" result))
	       (goto-char (point-min))
	       (or (re-search-forward
		    "Just a moment\\.+Pass phrase appears good\\. \\." nil t)
		   (error "!! Decryption failed."))
	       (setq result (buffer-substring (point) (point-max))))
	   (kill-buffer buffer))
	 result)))

(defun x-pgp-sig-lookup-keys (&optional regexp case pub)
  "Returns UserID/KeyID alist. [x-pgp-sig.el]"
  (let ((case-fold-search t)
	(keyring (if pub x-pgp-sig-pubring x-pgp-sig-secring))
	(re (format "^%s[\t ]+[0-9]+/\\([0-9a-f]+\\)[\t ]+[^\n\t ]+"
		    (if pub "pub" "sec")))
	process-connection-type
	buffer id result cell match)
    (or (file-exists-p keyring)
	(error "!! File not found; %s." keyring))
    (unwind-protect
	(save-excursion
	  (setq buffer (set-buffer
			(get-buffer-create x-pgp-sig-working-buffer))
		case-fold-search t
		buffer-read-only nil)
	  (erase-buffer)
	  (call-process
	   x-pgp-sig-pgp-program nil t nil
	   ;;; Suggested by Kazuyoshi Michishita <mit@unya.rim.or.jp>
	   "+verbose=1"
	   ;;;
	   "+language=en" "-kv" keyring)
	  (goto-char (point-min))
	  (when (re-search-forward
		 (format "^Key ring: '%s'\nType.+\n" (regexp-quote keyring))
		 nil t)
	    (while (re-search-forward re nil t)
	      (setq id (concat "0x" (match-string 1)))
	      (save-restriction
		(narrow-to-region
		 (match-end 0)
		 (or (and (re-search-forward "^[^\n\t ]" nil t)
			  (match-beginning 0))
		     (point-max)))
		(goto-char (point-min))
		(while (re-search-forward "[\t ]+\\(.+\\)$" nil t)
		  (setq result
			(append result (list (cons (match-string 1) id)))))
		(goto-char (point-max)))
	      )))
      (kill-buffer buffer))
    (if (stringp regexp)
	(progn
	  (setq case-fold-search case)
	  (while (setq cell (pop result))
	    (and (string-match regexp (car cell))
		 (setq match (append match (list cell)))))
	  match)
      result)))

(defun x-pgp-sig-build-message-to-sign ()
  "Extract fields and body to sign in current buffer. [x-pgp-sig.el]"
  (let ((case-fold-search t)
	(sign-fields (copy-sequence x-pgp-sig-sign-fields))
	(adjust x-pgp-sig-adjust-sign-fields-alist)
	(fields "")
	adj adj1 sub add existing-fields field body buffer)
    (save-excursion
      (save-restriction
	(x-pgp-sig-narrow-to-header t)
	(while (setq adj (pop adjust))
	  (setq adj1 (pop adj))
	  (goto-char (point-min))
	  (when (cond ((stringp adj1) (re-search-forward adj1 nil t))
		      ((fboundp adj1) (funcall adj1)))
	    (setq sub (cdr (assq 'sub adj))
		  add (cdr (assq 'add adj)))
	    (while sub
	      (setq sign-fields (delete (car sub) sign-fields))
	      (setq sub (cdr sub)))
	    (while add
	      (setq sign-fields (append sign-fields (list (car add)))
		    add (cdr add)))))
	(while (setq field (pop sign-fields))
	  (goto-char (point-min))
	  (and (re-search-forward (format "^%s:" (regexp-quote field)) nil t)
	       (setq fields (concat fields
				    (buffer-substring-no-properties
				     (match-beginning 0)
				     (or (and (re-search-forward
					       "^[^\n\t ]\\|^$" nil t)
					      (match-beginning 0))
					 (point-max))))
		     existing-fields (append existing-fields
					     (list field)))))
	(or existing-fields
	    (error "!! No fields to sign."))
	(goto-char (point-max))
	(widen)
	(setq body (buffer-substring-no-properties (point) (point-max)))
	(setq buffer
	      (set-buffer (get-buffer-create " *x-pgp-sig-msg-to-verify*")))
	(setq buffer-read-only nil)
	(erase-buffer)
	(insert body)
	(goto-char (point-min))
	(insert x-pgp-sig-signed-headers-field-name ": "
		(mapconcat 'identity existing-fields ",") "\n"
		fields "\n")
	(cons existing-fields buffer)))))

(defun x-pgp-sig-create-x-pgp-signature ()
  "To sign the message and create field body. [x-pgp-sig.el]"
  (let ((case-fold-search t)
	(msgs (x-pgp-sig-build-message-to-sign))
	(passphrase (x-pgp-sig-idea-decrypt-string
		     (cdr (assoc x-pgp-sig-current-signer
				 x-pgp-sig-passphrase-alist))))
	buffer result beg version)
    (or passphrase
	(setq passphrase (x-pgp-sig-set-signer)))
    (unwind-protect
	(save-excursion
	  (set-buffer (setq buffer (cdr msgs)))
	  (setq case-fold-search t)
	  (run-hooks 'x-pgp-sig-prepare-sign-hook)
	  (setq result
		(x-pgp-sig-process-buffer
		 (list "-fastu" x-pgp-sig-current-signer) passphrase))
	  (run-hooks 'x-pgp-sig-post-sign-hook)
	  (and result (error "!! %s" result))
	  (goto-char (point-min))
	  (or (re-search-forward
	       (format "^%s\n" (regexp-quote x-pgp-sig-signed-begin-line))
	       nil t)
	      (error "!! Sign failed."))
	  (setq beg (point))
	  (or (re-search-forward
	       (format "^%s$" (regexp-quote x-pgp-sig-signed-end-line)) nil t)
	      (error "!! Sign failed."))
	  (narrow-to-region beg (match-beginning 0))
	  (goto-char (point-min))
	  (setq version
		(or
		 (and (re-search-forward "^version:[\t ]*\\([^\n\t ]+\\)"
					 nil t)
		      (buffer-substring (match-beginning 1) (match-end 1)))
		 "0.0"))
	  (or (re-search-forward "\n\n\\([^\n\t ]\\)" nil t)
	      (error "!! Sign failed."))
	  (delete-region (point-min) (match-beginning 1))
	  (goto-char (point-min))
	  (insert version " "
		  (mapconcat 'identity (car msgs) ",")
		  "\n")
	  (while (not (eobp))
	    (insert "\t")
	    (end-of-line)
	    (forward-char 1))
	  (setq msgs (buffer-substring (point-min) (1- (point-max)))))
      (kill-buffer buffer))
    msgs))

(defun x-pgp-sig-build-message-to-verify (&optional buffer verbose)
  "Extract header and body to verify in specified buffer. [x-pgp-sig.el]"
  (and buffer
       (or (x-pgp-sig-buffer-live-p buffer)
	   (error "!! Buffer %s not exists." (prin1-to-string buffer)))
       (set-buffer (get-buffer buffer)))
  (let ((case-fold-search t)
	(warn-lo "Lines: arg out of range")
	(warn-ls "Lines: arg too small")
	(warn-co "Content-Length: arg out of range")
	(warn-cs "Content-Length: arg too small")
	(warn-m "Content-Length:,Lines: args mismatch")
	lines clength version beg end fields fieldsx field
	pgpsig pgphead body warn vbuff)
    (save-restriction
      (x-pgp-sig-narrow-to-header t)
      (and (re-search-forward "^lines:[\t ]+\\([0-9]+\\)" nil t)
	   (setq lines (string-to-int (match-string 1))))
      (goto-char (point-min))
      (and (re-search-forward
	    "^\\(x-\\)?content-length:[\t ]+\\([0-9]+\\)" nil t)
	   (setq clength (string-to-int (match-string 2))))
      (goto-char (point-min))
      (or (re-search-forward
	   (format "^%s:%s%s"
		   (regexp-quote x-pgp-sig-field-name)
		   "[\t ]+\\([^\n\t ,]+\\)[\t ]+\\([^\n\t ]+\\)"
		   "\\(\n\\([\t ]+.+\n\\)+\\|\\([\t ]+[^\n\t ]+\\)+\n\\)"
		   ) nil t)
	  (error "!! No correct \"%s\" field found." x-pgp-sig-field-name))
      (setq version (match-string 1)
	    beg (match-end 2)
	    end (match-end 0))
      (goto-char (match-beginning 2))
      (while (re-search-forward "[^\n\t ,]+" beg t)
	(setq fields (append fields (list (match-string 0)))))
      (setq fieldsx fields)
      (goto-char beg)
      (while (re-search-forward "[\t ]+\\([^\n\t ]+\\)" end t)
	(setq pgpsig (concat pgpsig (match-string 1) "\n")))
      (while (setq field (pop fields))
	(goto-char (point-min))
	(setq pgphead
	      (concat pgphead field
		      (or (and (re-search-forward
				(format "^\\(%s\\):\\([\t ]+.+\n\\)+"
					(regexp-quote field))
				nil t)
			       (buffer-substring-no-properties
				(match-end 1) (match-end 0)))
			  ": \n"))))
      (goto-char (point-max))
      (widen)
      (setq body
	    (buffer-substring-no-properties
	     (point)
	     (cond
	      ((and lines clength)
	       (if (> lines (count-lines (point) (point-max)))
		   (progn
		     (setq warn (cons warn-lo warn))
		     (if (> clength (- (point-max) (point)))
			 (progn
			   (setq warn (cons warn-co warn))
			   (point-max))
		       (prog1
			   (setq clength (+ (point) clength))
			 (or (eq clength (point-max))
			     (setq warn (cons warn-cs warn))))))
		 (if (> clength (- (point-max) (point)))
		     (progn
		       (setq warn (cons warn-co warn))
		       (next-line lines)
		       (progn
			 (or (eq (point) (point-max))
			     (setq warn (cons warn-ls warn)))
			 (point)))
		   (setq clength (+ (point) clength))
		   (or (eq clength (point-max))
		       (setq warn (cons warn-cs warn)))
		   (next-line lines)
		   (if (eq (point) clength)
		       (progn
			 (or (eq (point) (point-max))
			     (setq warn (cons warn-ls warn)))
			 (point))
		     (setq warn (cons warn-m warn))
		     (max (point) clength)))))
	      (lines
	       (if (> lines (count-lines (point) (point-max)))
		   (progn
		     (setq warn (cons warn-lo warn))
		     (point-max))
		 (next-line lines)
		 (progn
		   (or (eq (point) (point-max))
		       (setq warn (cons warn-ls warn)))
		   (point))))
	      (clength
	       (if (> clength (- (point-max) (point)))
		   (progn
		     (setq warn (cons warn-co warn))
		     (point-max))
		 (prog1
		     (setq clength (+ (point) clength))
		   (or (eq clength (point-max))
		       (setq warn warn-cs warn)))))
	      (t (point-max)))))
      (and verbose
	   warn
	   (mapcar (function (lambda (s)
			       (message "Warning! %s." s)
			       (sit-for 1)))
		   warn))
      (setq vbuff
	    (set-buffer (get-buffer-create " *x-pgp-sig-msg-to-verify*")))
      (setq buffer-read-only nil)
      (erase-buffer)
      (insert body)
      (or (eolp) (insert "\n"))
      (goto-char (point-min))
      (while (re-search-forward "^-" nil t)
	(replace-match "- -"))
      (goto-char (point-min))
      (insert x-pgp-sig-signed-msg-begin-line "\n"
	      "\n"
	      x-pgp-sig-signed-headers-field-name ": "
	      (mapconcat 'identity fieldsx ",") "\n"
	      pgphead
	      "\n")
      (goto-char (point-max))
      (insert "\n"
	      x-pgp-sig-signed-begin-line "\n"
	      "Version: " version "\n"
	      "\n"
	      pgpsig
	      x-pgp-sig-signed-end-line "\n")
      vbuff)))

(defun x-pgp-sig-put-field (field-name field-body)
  "Insert field without dragging faces. [x-pgp-sig.el]"
  (let ((case-fold-search t)
	(re (format "^%s:" (regexp-quote field-name)))
	(last (format "^\\(%s\\):"
		      (mapconcat 'identity x-pgp-sig-last-fields "\\|"))))
    (save-excursion
      (goto-char (point-min))
      (save-restriction
	(narrow-to-region
	 (point)
	 (or (and (re-search-forward
		   (format "^\\(%s\\)?\n" (regexp-quote mail-header-separator))
		   nil t)
		  (match-beginning 0))
	     (point-max)))
	(goto-char (point-min))
	(while (re-search-forward re nil t)
	  (delete-region
	   (match-beginning 0)
	   (or (and (re-search-forward "^[^\n\t ]\\|^$" nil t)
		    (match-beginning 0))
	       (point-max))))
	(goto-char (point-min))
	(if (re-search-forward last nil t)
	    (goto-char (match-beginning 0))
	  (goto-char (point-max)))
	(if (bobp)
	    (insert field-name ": " field-body "\n")
	  (backward-char 1)
	  (insert "\n" field-name ": " field-body))))))


;;; Commands and functions to be autoloaded.

;;;###autoload
(defun x-pgp-sig-sign (&optional force)
  "Add X-PGP signature. [x-pgp-sig.el]"
  (when (or force
	    x-pgp-sig-enable-sign
	    (if (fboundp x-pgp-sig-add-signature-judging-function)
		(funcall x-pgp-sig-add-signature-judging-function)
	      x-pgp-sig-always-sign))
    (let ((case-fold-search t)
	  end clength signature)
      (save-excursion
	(save-restriction
	  (widen)
	  (goto-char (point-min))
	  (when (re-search-forward
		 (format "^\\(%s\\)?\n" (regexp-quote mail-header-separator))
		 nil t)
	    (setq end (match-beginning 0)
		  clength (- (point-max) (match-end 0)))
	    (goto-char (point-min))
	    (or (zerop clength)
		(re-search-forward "^control:[\t ]*cancel" end t)
		(progn
		  (and x-pgp-sig-approved-field-body
		       (re-search-forward "^control:" end t)
		       (x-pgp-sig-put-field "Approved"
					    x-pgp-sig-approved-field-body))
		  (and x-pgp-sig-add-content-length-field
		       (x-pgp-sig-put-field "Content-Length"
					    (int-to-string clength)))
		  (and x-pgp-sig-x-info-field-body
		       (x-pgp-sig-put-field "X-Info"
					    x-pgp-sig-x-info-field-body))
		  (when
		      (setq signature
			    (condition-case err
				(progn
				  (message "Signing the message...")
				  (prog1
				      (x-pgp-sig-create-x-pgp-signature)
				    (message "Signing the message...done")))
			      (error
			       (progn
				 (message (if (and (listp err)
						   (stringp (car (cdr err))))
					      (car (cdr err))
					    (prin1-to-string err)))
				 nil))))
		    (x-pgp-sig-put-field x-pgp-sig-field-name signature)
		    (run-hooks 'x-pgp-sig-sign-hook)))))))))
  (setq x-pgp-sig-enable-sign nil))

;;;###autoload
(defun x-pgp-sig-set-signer (&optional signer passphrase)
  "Set PGP signer and passphrase. [x-pgp-sig.el]"
  (interactive)
  (let ((signer (or signer
		    (if (stringp x-pgp-sig-current-signer)
			x-pgp-sig-current-signer
		      (user-login-name))))
	(keys (or x-pgp-sig-secret-key-alist
		  (setq x-pgp-sig-secret-key-alist (x-pgp-sig-lookup-keys))
		  (error "!! No keys in %s." x-pgp-sig-secring)))
	(interactive (interactive-p))
	case-fold-search orig id key result)
    (setq orig signer)
    (and (string-match "^[^\n\t <>]+$" signer)
	 (setq case-fold-search t))
    (while (setq key (pop keys))
      (and (string-match signer (car key))
	   (setq id (car key)
		 keys nil)))
    (setq signer
	  (if interactive
	      (completing-read "Enter the key's user ID: "
			       x-pgp-sig-secret-key-alist nil t id)
	    (if id
		id
	      (error "!! No secret keys for \"%s\" found." orig))))
    (setq key (cdr (assoc signer x-pgp-sig-secret-key-alist)))
    (unless (or passphrase (assoc signer x-pgp-sig-passphrase-alist))
      (setq interactive t)
      (setq passphrase (x-pgp-sig-read-noecho
			(format "PGP passphrase for %s (%s): " signer key)))
      (and (zerop (length passphrase)) (error "!! Empty passphrase.")))
    (when passphrase
      (when interactive
	(message "Confirming passphrase...")
	(let ((buffer (get-buffer-create x-pgp-sig-working-buffer))
	      pair)
	  (unwind-protect
	      (save-excursion
		(set-buffer buffer)
		(setq buffer-read-only nil)
		(erase-buffer)
		(setq result
		      (x-pgp-sig-process-buffer
		       (list "-fastu" signer) passphrase))
		(if result
		    (error "!! %s" result)
		  (message "Confirming passphrase...done")))
	    (kill-buffer buffer))))
      (setq x-pgp-sig-passphrase-alist
	    (put-alist signer
		       (progn (message "Encrypting passphrase...")
			      (x-pgp-sig-idea-encrypt-string passphrase))
		       x-pgp-sig-passphrase-alist)))
    (setq x-pgp-sig-current-signer signer)
    (prog1
	(or passphrase
	    (progn
	      (message "Decrypting passphrase...")
	      (x-pgp-sig-idea-decrypt-string
	       (cdr (assoc signer x-pgp-sig-passphrase-alist)))))
      (message "The signer and the passphrase have been registered."))))

;;;###autoload
(defun x-pgp-sig-set-signer-maybe ()
  "Set PGP signer and passphrase if necessary. [x-pgp-sig.el]"
  (interactive)
  (or (stringp x-pgp-sig-default-signer)
      (setq x-pgp-sig-default-signer
	    (format "%s <%s>" (user-full-name) x-pgp-sig-from)))
  (and (or
	(not (string-equal x-pgp-sig-current-signer x-pgp-sig-default-signer))
	(not (assoc x-pgp-sig-current-signer x-pgp-sig-passphrase-alist)))
       (condition-case err
	   (call-interactively 'x-pgp-sig-set-signer)
	 (error (message "%s" (if (and (listp err) (stringp (car (cdr err))))
				  (car (cdr err))
				(prin1-to-string err)))))))

;;;###autoload
(defun x-pgp-sig-switch-signing (&optional switch)
  "Judging whether to add X-PGP signature. [x-pgp-sig.el]"
  (interactive
   (list (y-or-n-p
	  (format "Do you want to add %s field? " x-pgp-sig-field-name))))
  (let ((msg "This message will %sbe signed."))
    (if (setq x-pgp-sig-enable-sign (not (not switch)))
	(progn
	  (and (boundp 'mime-editor/pgp-processing)
	       (eq mime-editor/pgp-processing 'sign)
	       (setq mime-editor/pgp-processing nil))
	  (setq msg (format msg "")))
      (setq msg (format msg "not ")))
    (and (interactive-p) (message msg))))

;;;###autoload
(defun x-pgp-sig-verify (&optional verbose)
  "Verify X-PGP signature in current message. [x-pgp-sig.el]"
  (interactive "P")
  (let ((method (assq major-mode x-pgp-sig-verify-method-alist))
	(delete x-pgp-sig-delete-last-empty-line-while-verifying)
	(re (format "\\(\n+\\)%s" (regexp-quote x-pgp-sig-signed-begin-line)))
	(count 0)
	prepare buffer post vbuff vbuf2 newlines good bad pgp-id)
    (and method
	 (setq method (cdr method)
	       prepare (car method)
	       buffer (car (cdr method))
	       post (car (cdr (cdr method)))))
    (message "Verifying signature...")
    (unwind-protect
	(progn
	  (and prepare
	       (let ((gnus-show-all-headers t)
		     gnus-show-mime)
		 (funcall prepare)))
	  (setq buffer (condition-case ()
			   (if (fboundp buffer)
			       (funcall buffer)
			     (symbol-value buffer))
			 (error nil)))
	  (or (x-pgp-sig-buffer-live-p buffer)
	      (setq buffer (current-buffer)))
	  (save-excursion
	    (setq vbuff
		  (set-buffer
		   (x-pgp-sig-build-message-to-verify buffer verbose)))
	    (goto-char (point-min))
	    (setq newlines
		  (or (and (re-search-forward re nil t)
			   (length (match-string 1)))
		      0))
	    (setq vbuf2
		  (set-buffer
		   (get-buffer-create " *x-pgp-sig-msg-to-verify-2*")))
	    (setq buffer-read-only nil)
	    (while (or delete (not (or good bad)))
	      (erase-buffer)
	      (insert-buffer-substring vbuff)
	      (and delete
		   (> count 0)
		   (> newlines (1+ count))
		   (goto-char (point-min))
		   (re-search-forward re nil t)
		   (goto-char (match-beginning 0))
		   (message "[%d]Verifying signature..." count)
		   (delete-char count))
	      (or (> newlines (+ 2 count))
		  (setq delete nil))
	      (x-pgp-sig-process-buffer '("-f") nil vbuf2)
	      (incf count)
	      (goto-char (point-min))
	      (and (re-search-forward x-pgp-sig-good-signature-regexp nil t)
		   (setq good (match-string 0)
			 delete nil))
	      (goto-char (point-min))
	      (if (re-search-forward x-pgp-sig-expected-re nil t)
		  (progn
		    (setq pgp-id (concat "0x"
					 (buffer-substring-no-properties
					  (match-beginning 1) (match-end 1))))
		    (or
		     (and (or (boundp 'mc-pgp-always-fetch)
			      (condition-case ()
				  (load-library "mc-pgp"))
			      (error nil))
			  (not (eq mc-pgp-always-fetch 'never))
			  (or mc-pgp-always-fetch
			      (y-or-n-p
			       (format
				"Key %s not found; attempt to fetch? "
				pgp-id)))
			  (mc-pgp-fetch-key (cons pgp-id nil))
			  (setq count 0))
		     (setq bad (format "!! Key %s not found." pgp-id))))
		(goto-char (point-min))
		(setq bad (concat "!! "
				  (or (and (re-search-forward
					    x-pgp-sig-bad-signature-regexp
					    nil t)
					   (match-string 0))
				      "Bad signature.")))))))
      (and vbuff (kill-buffer vbuff))
      (and vbuf2 (kill-buffer vbuf2))
      (and post (funcall post))
      (and (x-pgp-sig-buffer-live-p x-pgp-sig-msg-to-verify-buffer)
	   (kill-buffer x-pgp-sig-msg-to-verify-buffer)))
    (decf count)
    (message "%s"
	     (if good
		 (if (zerop count)
		     good
		   (format "[%d]%s" count good))
	       bad))))

(defun x-pgp-sig-add-signature-y-or-n-p ()
  (y-or-n-p "Do you want to add X-PGP-Sig field? "))


;;; GNUS, Gnus
(defun x-pgp-sig-find-gnus-article-buffer ()
  "Find article buffer may includes X-PGP-Sig field. [x-pgp-sig.el]"
  (or (and (boundp 'gnus-original-article-buffer)
	   (x-pgp-sig-buffer-live-p gnus-original-article-buffer))
      (and (boundp 'gnus-article-buffer)
	   (x-pgp-sig-buffer-live-p gnus-article-buffer))
      (and (boundp 'gnus-Article-buffer)
	   (x-pgp-sig-buffer-live-p gnus-Article-buffer))))

;;; MH-E
(defun x-pgp-sig-mh-get-article ()
  "Get current message in article buffer. [x-pgp-sig.el]"
  (let ((file (concat mh-folder-filename (mh-get-msg-num t)))
	; NEmacs
	kanji-flag
	; Mule
	(input-coding-system x-pgp-sig-*noconv*)
	; >= XEmacs-20.1b6
	(coding-system-for-read x-pgp-sig-*noconv*))
    (save-excursion
      (setq x-pgp-sig-msg-to-verify-buffer
	    (set-buffer (get-buffer-create " *x-pgp-sig-tmp-article*")))
      (setq buffer-read-only nil)
      (erase-buffer)
      (insert-file-contents file))))

;;; Mew
(defun x-pgp-sig-mew-get-article ()
  "Get current message in article buffer. [x-pgp-sig.el]"
  (let ((file
	 (mew-expand-file-name (mew-summary-message-number)
			       (mew-summary-folder-name)))
	; Mule-2.x
	(input-coding-system mew-cs-spool)
	; >= XEmacs-20.1b6
	(coding-system-for-read mew-cs-spool))
    (save-excursion
      (setq x-pgp-sig-msg-to-verify-buffer
	    (set-buffer (get-buffer-create " *x-pgp-sig-tmp-article*")))
      (setq buffer-read-only nil)
      (erase-buffer)
      (insert-file-contents file))))

(defun x-pgp-sig-mew-encode-message-header-with-semi ()
  "Encode the message header to MIME format with SEMI. [x-pgp-sig.el]"
  (require 'eword-encode)
  (eword-encode-header))

(defun x-pgp-sig-mew-encode-message-header-with-tm ()
  "Encode the message header to MIME format with tm. [x-pgp-sig.el]"
  (require 'tm-ew-e)
  (mime/encode-message-header))

(defun x-pgp-sig-mew-encode-message-header-with-toMime ()
  "Encode the message header to MIME format with toMime. [x-pgp-sig.el]"
  (save-excursion
    (save-restriction
      (x-pgp-sig-narrow-to-header)
      (let (
	    ; Mule-2.x
	    (input-coding-system mew-cs-noconv)
	    (output-coding-system mew-cs-draft)
	    ; XEmacs-20.{1,2}
	    (coding-system-for-read mew-cs-noconv)
	    (coding-system-for-write mew-cs-draft)
	    )
	(apply 'call-process-region (point-min) (point-max)
	       x-pgp-sig-toMime-program t t nil
	       x-pgp-sig-toMime-options)))))

(defun x-pgp-sig-mew-encode-message ()
  "Encode the message to the network code. [x-pgp-sig.el]"
  (make-local-variable 'mail-header-separator)
  (setq mail-header-separator mew-header-separator)
  (save-excursion
    (save-restriction
      (widen)
      (and x-pgp-sig-mew-encode-message-header-function
	   (funcall x-pgp-sig-mew-encode-message-header-function))
      (goto-char (point-min))
      (and (re-search-forward "\\cj+" nil t)
	   (mew-cs-encode-region (point-min) (point-max) mew-cs-draft)))))


(provide 'x-pgp-sig)
(run-hooks 'x-pgp-sig-load-hook)
;;; x-pgp-sig.el ends here.
